// nanoCoder.cpp
// 

#include "WavFile.h" // I/O wave file

#include <stdexcept>
#include <iostream>
#include <cmath>

using namespace std;

const double TWOPI = M_PI * 2.;

void makeWindow (float* out, int N, float a0, float a1, float a2) {
    // .5, .5, 0     --> hanning
    // .54, .46, 0   --> hamming
    // .42, .5, 0.08 --> blackman
    for (int i = 0; i < N; ++i) {
        out[i] = a0 - a1 * cos ((TWOPI * (float) i) / (N - 1)) + a2 * 
        	cos ((2 * TWOPI * (float) i) / (N - 1)); // hamming, hann or blackman
    }
}

void fft (float *fftBuffer, long fftFrameSize, long sign) {
    float wr, wi, arg, *p1, *p2, temp;
    float tr, ti, ur, ui, *p1r, *p1i, *p2r, *p2i;
    long i, bitm, j, le, le2, k;

    for (i = 2; i < 2*fftFrameSize-2; i += 2) {
        for (bitm = 2, j = 0; bitm < 2*fftFrameSize; bitm <<= 1) {
            if (i & bitm) j++;
            j <<= 1;
        }
        if (i < j) {
            p1 = fftBuffer+i;
            p2 = fftBuffer+j;
            temp = *p1;
            *(p1++) = *p2;
            *(p2++) = temp;
            temp = *p1;
            *p1 = *p2;
            *p2 = temp;
        }
    }
    for (k = 0, le = 2; k < (long)(log(fftFrameSize)/log(2.)); k++) {
        le <<= 1;
        le2 = le>>1;
        ur = 1.0;
        ui = 0.0;
        arg = M_PI / (le2>>1);
        wr = cos(arg);
        wi = sign*sin(arg);
        for (j = 0; j < le2; j += 2) {
            p1r = fftBuffer+j;
            p1i = p1r+1;
            p2r = p1r+le2;
            p2i = p2r+1;
            for (i = j; i < 2*fftFrameSize; i += le) {
                tr = *p2r * ur - *p2i * ui;
                ti = *p2r * ui + *p2i * ur;
                *p2r = *p1r - tr;
                *p2i = *p1i - ti;
                *p1r += tr;
                *p1i += ti;
                p1r += le;
                p1i += le;
                p2r += le;
                p2i += le;
            }
            tr = ur*wr - ui*wi;
            ui = ur*wi + ui*wr;
            ur = tr;
        }
    }
}

// rect to amp-freq
void convert (const float* cbuffer, float* amp, float* freq, float* oldPhi,
              int N, int hop, float R) {
    float osamp = N / hop;
    float freqPerBin = R / (float) N;
    float expct = TWOPI * (float) hop / (float) N;
    for (int i = 0; i < N; ++i) {
        amp[i] = 2. * sqrt (cbuffer[2 * i] * cbuffer[2 * i]
                            + cbuffer[2 * i + 1] * cbuffer[2 * i + 1]);

        float phase = atan2 (cbuffer[2 * i + 1], cbuffer[2 * i]);
        float tmp = phase - oldPhi[i];
        oldPhi[i] = phase;
        tmp -= (float) i * expct;
        int qpd = (int) (tmp / M_PI);
        if (qpd >= 0) qpd += qpd & 1;
        else qpd -= qpd & 1;
        tmp -= M_PI * (float) qpd;
        tmp = osamp * tmp / (2. * M_PI);
        tmp = (float) i * freqPerBin + tmp * freqPerBin;
        freq[i] = tmp;
    }
}

// amp-freq to rect
void unconvert (const float* amp, const float* freq, float* oldPhi,
                float* cbuffer, int N, int hop, float R) {
    float osamp = N / hop;
    float freqPerBin = R / (float) N;
    float expct = TWOPI * (float) hop / (float) N;

    for (int i = 0; i < N; ++i) {
        float tmp = freq[i];

        tmp -= (float)i * freqPerBin;
        tmp /= freqPerBin;
        tmp = TWOPI * tmp / osamp;
        tmp += (float) i * expct;
        oldPhi[i] += tmp;
        
        float phase = (oldPhi[i]); // - oldPhi[i - 1] - oldPhi[i + 1]);

        cbuffer[2 * i] = amp[i] * cos (phase);
        cbuffer[2 * i + 1] = amp[i] * sin (phase);
    }
}

void process (float* window, float* workspace, 
	float* amp, float* freq, float* old_phi, float* old_phisyn,
	float* ampsyn, float* freqsyn, float pitch, float threshold,
    int fftlen, int hop, int outhop, int sr) {

	 fft (workspace, fftlen, -1); // forward fft
     convert (workspace, amp, freq, old_phi, fftlen, hop, sr);
    
    // anti-noise
    for (int i = 0; i < fftlen; ++i) {
        if (amp[i] < threshold) amp[i] = 0;
    }

    // pitch tranpsosition                      
    memset (ampsyn, 0, sizeof (float) * fftlen);
    memset (freqsyn, 0, sizeof (float) * fftlen);
    int hspect = fftlen * pitch;
    for (int i = 0; i < fftlen; ++i) {
        int idx = ((int) i / pitch);
        if (idx < hspect) {
            ampsyn[i] += amp[idx];
            freqsyn[i] = freq[idx] * pitch;
        }
    }
    
    unconvert (ampsyn, freqsyn, old_phisyn, workspace, fftlen, outhop, sr);
	fft (workspace, fftlen, 1); // inverse fft
}

int main (int argc, char* argv[]) {
    cout << "[nanoCoder, ver. 0.1]" << endl << endl;
    cout << "minimalistic phase-vocoder" << endl;
    cout << "(c) 2014, www.carminecella.com" << endl << endl;

	try {
		if (argc < 8 || argc > 9) {
			cout << "error: syntax is 'nanoCoder in.wav out.wav pshift tstretch thr fft olap [scale]'" << endl;
			return 0;
		}

        // parameters
		WavInFile input (argv[1]);
		int sr = input.getSampleRate ();
		int nchan = input.getNumChannels ();
		int samples = input.getNumSamples ();
		int bits = input.getNumBits ();

		if (bits != 16) {
			cout << "error: 16 bit mono files supported only" << endl;
			return 0;
		}

		WavOutFile output (argv[2], sr, bits, nchan);

		float ps = atof (argv[3]);
		if (ps < 0.125 || ps > 4) {
			cout << "error: invalid shift ratio [.124 : 4]" << endl;
			return 0;
		}
        
        float tstretch = atof (argv[4]);
        if (tstretch < 0.125 || ps > 4) {
            cout << "error: invalid stretch ratio [.124 : 4]" << endl;
            return 0;
        }        
        
        float thr = atof (argv[5]);

        int fftlen = atol (argv[6]);
        if ((((~fftlen + 1) & fftlen) != fftlen) || fftlen < 2) { // check pow 2
            cout << "error: fft size must be a power of two" << endl;
            return 0;
        }
        
        int olap = atol (argv[7]);
        int outhop = fftlen / olap;
        int hopsize = (int) ((float) outhop / tstretch);

        float scale = 1;
        if (argc == 9) {
            scale = atof (argv[8]);
        }

		// memory allocation
		float* buffer = new float[fftlen];
		float* window = new float[fftlen];
		makeWindow (window, fftlen, .5, .5, 0);

		float* amp = new float[fftlen];
		float* freq = new float[fftlen];
		float* ampsyn = new float[fftlen];
		float* freqsyn = new float[fftlen];		
		float* old_phi = new float[fftlen];
		memset (old_phi, 0, sizeof (float) * fftlen);
        float* old_phisyn = new float[fftlen];
        memset (old_phisyn, 0, sizeof (float) * fftlen);

		float* workspace = new float[fftlen * 2];

		float* input_data = new float[nchan * samples + fftlen]; // safety
		int outsamples = (int) (nchan * samples * tstretch + fftlen); // safety
		float* output_data = new float[outsamples];
		memset (output_data, 0, sizeof (float) * outsamples);

        // heuristic for normalization
        float norm = (1. / (fftlen * olap));
        if (tstretch != 1) norm *= (tstretch * 1.2);
        if (ps > 1) norm = pow (norm, 1. / (ps * .6));
        else if (ps < 1) norm = pow (norm, ps * 1.75);
        norm *= scale;

        // printing info
        cout << "fft       : " << fftlen << endl;
        cout << "input hop : " << hopsize << endl;
        cout << "output hop: " << outhop << endl;
        cout << "stretch   : " << tstretch << "x" << endl;
        cout << "shift     : " << ps << "x" << endl;
        cout << "threshold : " << thr << endl;
        cout << "scale     : " << scale << endl << endl;

        cout << "reading data..."; cout.flush ();
		input.read (input_data, samples * nchan);
        cout << "done" << endl;
		
        cout << "processing....."; cout.flush ();
        for (int j = 0; j < nchan; ++j) {
            int pointer = 0;
            int wpointer = 0;

    		while (pointer < samples) {
                for (int k = pointer; k < pointer + fftlen; ++k) {
                    workspace[2 * (k - pointer)] = input_data[nchan * k + j] 
                    * window[k - pointer];
                    workspace[2 * (k - pointer) + 1] = 0;
                }

    			process (
                    window, workspace, 
                    amp, freq, old_phi, 
                    old_phisyn, ampsyn, freqsyn,
    				ps, thr,
                    fftlen, hopsize, outhop, sr);
    
                 // overlapp-add
                 for (int k = (int) wpointer; k < (int) (wpointer + fftlen); ++k) {
                    output_data[nchan * k + j] += (workspace[2 * (k - wpointer)]
                        * window[k - wpointer] * norm);
                 }           

    			pointer += hopsize;
                wpointer += outhop;
            }
		}
        cout << "done" << endl;

        cout << "writing data..."; cout.flush ();
		output.write (output_data, outsamples);
        cout << "done" << endl << endl;

		// memory deallocation
		delete [] buffer;
		delete [] amp;
		delete [] freq;
		delete [] ampsyn;
		delete [] freqsyn;		
		delete [] old_phi;
        delete [] old_phisyn;
		delete [] workspace;
		delete [] input_data;
		delete [] output_data;
		delete [] window;

	} catch (exception& e) {
		cout << e.what () << endl;
	}

	return 0;
}

// EOF
